TaxiService = {};

local AllVehicleCameras = {}

local StuckTrys = 0;

local TaxiVehicle = nil;
local AIDriveToJob = nil;
local TaxiState = 0;

local CoursePickupX = nil;
local CoursePickupZ = nil;
local DestinationX = nil;
local DestinationZ = nil;

local ExitTaxiVehicle = nil;
local ExitAIDriveToJob = nil;


function TaxiService:DistanceFrom(x1,y1,x2,y2) return math.sqrt((x2 - x1) ^ 2 + (y2 - y1) ^ 2) end

function TaxiService:UpdateCamera(CameraToUpdate)
	Mainobj = CameraToUpdate
	local dt = 1;
    local target = Mainobj.zoomTarget
    if Mainobj.zoomLimitedTarget >= 0 then
        target = math.min(Mainobj.zoomLimitedTarget, Mainobj.zoomTarget)
    end
    Mainobj.zoom = target + ( math.pow(0.99579, dt) * (Mainobj.zoom - target) )

    --
    if Mainobj.lastInputValues.upDown ~= 0 then
        local value = Mainobj.lastInputValues.upDown * g_gameSettings:getValue(GameSettings.SETTING.CAMERA_SENSITIVITY)
        Mainobj.lastInputValues.upDown = 0
        value = g_gameSettings:getValue("invertYLook") and -value or value

        if Mainobj.isRotatable then
            if Mainobj.isActivated and not g_gui:getIsGuiVisible() then
                if Mainobj.limitRotXDelta > 0.001 then
                    Mainobj.rotX = math.min(Mainobj.rotX - value, Mainobj.rotX)
                elseif Mainobj.limitRotXDelta < -0.001 then
                    Mainobj.rotX = math.max(Mainobj.rotX - value, Mainobj.rotX)
                else
                    Mainobj.rotX = Mainobj.rotX - value
                end

                if Mainobj.limit then
                    Mainobj.rotX = math.min(Mainobj.rotMaxX, math.max(Mainobj.rotMinX, Mainobj.rotX))
                end
            end
        end
    end

    if Mainobj.lastInputValues.leftRight ~= 0 then
        local value = Mainobj.lastInputValues.leftRight * g_gameSettings:getValue(GameSettings.SETTING.CAMERA_SENSITIVITY)
        Mainobj.lastInputValues.leftRight = 0

        if Mainobj.isRotatable then
            if Mainobj.isActivated and not g_gui:getIsGuiVisible() then
                Mainobj.rotY = Mainobj.rotY - value
            end
        end
    end

    --
    if g_gameSettings:getValue("isHeadTrackingEnabled") and isHeadTrackingAvailable() and Mainobj.allowHeadTracking and Mainobj.headTrackingNode ~= nil then
        local tx,ty,tz = getHeadTrackingTranslation()
        local pitch,yaw,roll = getHeadTrackingRotation()
        if pitch ~= nil then
            local camParent = getParent(Mainobj.cameraNode)
            local ctx,cty,ctz
            local crx,cry,crz
            if camParent ~= 0 then
                ctx, cty, ctz = localToLocal(Mainobj.headTrackingNode, camParent, tx, ty, tz)
                crx, cry, crz = localRotationToLocal(Mainobj.headTrackingNode, camParent, pitch,yaw,roll)
            else
                ctx, cty, ctz = localToWorld(Mainobj.headTrackingNode, tx, ty, tz)
                crx, cry, crz = localRotationToWorld(Mainobj.headTrackingNode, pitch,yaw,roll)
            end

            setRotation(Mainobj.cameraNode, crx, cry, crz)
            setTranslation(Mainobj.cameraNode, ctx, cty, ctz)
        end
    else
        Mainobj:updateRotateNodeRotation()

        if Mainobj.limit then
            -- adjust rotation to avoid clipping with terrain
            if Mainobj.isRotatable and ((Mainobj.useWorldXZRotation == nil and g_gameSettings:getValue("useWorldCamera")) or Mainobj.useWorldXZRotation) then
                local numIterations = 4
                for _=1, numIterations do
                    local transX, transY, transZ = Mainobj.transDirX*Mainobj.zoom, Mainobj.transDirY*Mainobj.zoom, Mainobj.transDirZ*Mainobj.zoom
                    local x,y,z = localToWorld(getParent(Mainobj.cameraPositionNode), transX, transY, transZ)

                    local terrainHeight = DensityMapHeightUtil.getHeightAtWorldPos(x,0,z)

                    local minHeight = terrainHeight + 0.9
                    if y < minHeight then
                        local h = math.sin(Mainobj.rotX)*Mainobj.zoom
                        local h2 = h-(minHeight-y)
                        Mainobj.rotX = math.asin(MathUtil.clamp(h2/Mainobj.zoom, -1, 1))
                        Mainobj:updateRotateNodeRotation()
                    else
                        break
                    end
                end
            end

            -- adjust zoom to avoid collision with objects
            if Mainobj.allowTranslation then

                Mainobj.limitRotXDelta = 0
                local hasCollision, collisionDistance, nx,ny,nz, normalDotDir = Mainobj:getCollisionDistance()
                if hasCollision then
                    local distOffset = 0.1
                    if normalDotDir ~= nil then
                        local absNormalDotDir = math.abs(normalDotDir)
                        distOffset = MathUtil.lerp(1.2, 0.1, absNormalDotDir*absNormalDotDir*(3-2*absNormalDotDir))
                    end
                    collisionDistance = math.max(collisionDistance-distOffset, 0.01)
                    Mainobj.disableCollisionTime = g_currentMission.time+400
                    Mainobj.zoomLimitedTarget = collisionDistance
                    if collisionDistance < Mainobj.zoom then
                        Mainobj.zoom = collisionDistance
                    end
                    if Mainobj.isRotatable and nx ~= nil and collisionDistance < Mainobj.transMin then
                        local _,lny,_ = worldDirectionToLocal(Mainobj.rotateNode, nx,ny,nz)
                        if lny > 0.5 then
                            Mainobj.limitRotXDelta = 1
                        elseif lny < -0.5 then
                            Mainobj.limitRotXDelta = -1
                        end
                    end
                else
                    if Mainobj.disableCollisionTime <= g_currentMission.time then
                        Mainobj.zoomLimitedTarget = -1
                    end
                end
            end

        end
        Mainobj.transX, Mainobj.transY, Mainobj.transZ = Mainobj.transDirX*Mainobj.zoom, Mainobj.transDirY*Mainobj.zoom, Mainobj.transDirZ*Mainobj.zoom
        setTranslation(Mainobj.cameraPositionNode, Mainobj.transX, Mainobj.transY, Mainobj.transZ)

        if Mainobj.positionSmoothingParameter > 0 then

            local interpDt = g_physicsDt

            if Mainobj.vehicle.spec_rideable ~= nil then
                interpDt = Mainobj.vehicle.spec_rideable.interpolationDt
            end

            if g_server == nil then
                -- on clients, we interpolate the vehicles with dt, thus we need to use the same for camera interpolation
                interpDt = dt
            end
            if interpDt > 0 then
                local xlook,ylook,zlook = getWorldTranslation(Mainobj.rotateNode)
                local lookAtPos = Mainobj.lookAtPosition
                local lookAtLastPos = Mainobj.lookAtLastTargetPosition
                lookAtPos[1],lookAtPos[2],lookAtPos[3] = Mainobj:getSmoothed(Mainobj.lookAtSmoothingParameter, lookAtPos[1],lookAtPos[2],lookAtPos[3], xlook,ylook,zlook, lookAtLastPos[1],lookAtLastPos[2],lookAtLastPos[3], interpDt)
                lookAtLastPos[1],lookAtLastPos[2],lookAtLastPos[3] = xlook,ylook,zlook

                local x,y,z = getWorldTranslation(Mainobj.cameraPositionNode)
                local pos = Mainobj.position
                local lastPos = Mainobj.lastTargetPosition
                pos[1],pos[2],pos[3] = Mainobj:getSmoothed(Mainobj.positionSmoothingParameter, pos[1],pos[2],pos[3], x,y,z, lastPos[1],lastPos[2],lastPos[3], interpDt)
                lastPos[1],lastPos[2],lastPos[3] = x,y,z

                local upx, upy, upz = localDirectionToWorld(Mainobj.rotateNode, Mainobj:getTiltDirectionOffset(), 1, 0)
                local up = Mainobj.upVector
                local lastUp = Mainobj.lastUpVector
                up[1],up[2],up[3] = Mainobj:getSmoothed(Mainobj.positionSmoothingParameter, up[1],up[2],up[3], upx, upy, upz, lastUp[1],lastUp[2],lastUp[3], interpDt)
                lastUp[1],lastUp[2],lastUp[3] = upx, upy, upz

                Mainobj:setSeparateCameraPose()
            end
        end

    end

end

function TaxiService:RequestLocation()
    local Options = {};
    local Arguments = {};
    local Types = {};

    for i = 1, #g_fieldManager:getFields() do
        Options[i] = string.format(g_i18n:getText("fieldJob_number"), i)
        Arguments[i] = i;
        Types[i] = "FIELD";
    end

    for i=#g_currentMission.placeableSystem.placeables, 1, -1 do
        local Placeable = g_currentMission.placeableSystem.placeables[i];

        Options[table.getn(Options) + 1] = Placeable:getName();
        Arguments[table.getn(Arguments) + 1] = Placeable;
        Types[table.getn(Types) + 1] = "PLACEABLE";
    end

    g_gui:showOptionDialog({text = "",title = "",options = Options, args = Arguments,
    callback = function(sel,arg)

        local Tx = 0;
        local Tz = 0;

        if Types[sel] == "PLACEABLE" then
            Tx = Arguments[sel].position.x;
            Tz = Arguments[sel].position.z;
        end;

        if Types[sel] == "FIELD" then
            local Field = g_fieldManager:getFieldByIndex(Arguments[sel]);

            Tx = Field.posX;
            Tz = Field.posZ;
        end;

        DestinationX = Tx;
        DestinationZ = Tz;

        Options = nil;
        Types = nil;
        Arguments = nil;

        AIDriveToJob = g_currentMission.aiJobTypeManager:createJob(g_currentMission.aiJobTypeManager:getJobTypeIndexByName("GOTO"))
		
        AIDriveToJob.vehicleParameter:setVehicle(TaxiVehicle);
		AIDriveToJob.positionAngleParameter:setAngle(0);
		AIDriveToJob.positionAngleParameter:setPosition(Tx,Tz);

        AIDriveToJob.driveToTask:setVehicle(TaxiVehicle);
        AIDriveToJob.driveToTask:setTargetDirection(0, 0);
        AIDriveToJob.driveToTask:setTargetPosition(Tx,Tz);

        g_currentMission.aiSystem:startJob(AIDriveToJob, -1);
        TaxiState = 2;
    end});
end;

function TaxiService:UpdateCameras()
    for i = 1, table.getn(AllVehicleCameras) do
		if AllVehicleCameras[i].isActivated == true and AllVehicleCameras[i].vehicle == TaxiVehicle then
			TaxiService:UpdateCamera(AllVehicleCameras[i]);
		end;
	end
end;

function TaxiService:SpawnTaxiVehicle(StartPosition, IsExit)

    --local xmlFilename = "data/vehicles/mahindra/retriever/retriever.xml";
    local xmlFilename = "data/vehicles/lizard/pickup2017/pickup2017.xml";

	local SpawnCallback = function(_, vehicle, vehicleLoadState, arguments)	
        if IsExit == nil then TaxiVehicle = vehicle; end;
        if IsExit ~= nil then ExitTaxiVehicle = vehicle; end;
	end;

	VehicleLoadingUtil.loadVehicle(xmlFilename, StartPosition, true, 0, Vehicle.PROPERTY_STATE_OWNED, -1, {}, nil, SpawnCallback, nil, nil, true, false);
end

function TaxiService:update()	

    TaxiService:UpdateCameras();

    if TaxiState == 3 and ExitTaxiVehicle ~= nil and ExitAIDriveToJob ~= nil and not Input.isKeyPressed(Input.KEY_g) then

        TaxiState = 0;
        CoursePickupX = nil;
        CoursePickupZ = nil;
        AIDriveToJob = nil;
        TaxiVehicle = nil;

        --print("SUPEREXIT")
    end

    if TaxiState == 3 and ExitTaxiVehicle ~= nil then

        --print("STATE 3")

        ExitAIDriveToJob = g_currentMission.aiJobTypeManager:createJob(g_currentMission.aiJobTypeManager:getJobTypeIndexByName("GOTO"))
		
        ExitAIDriveToJob.vehicleParameter:setVehicle(ExitTaxiVehicle);
		ExitAIDriveToJob.positionAngleParameter:setAngle(0);
		ExitAIDriveToJob.positionAngleParameter:setPosition(-100,-100);

        ExitAIDriveToJob.driveToTask:setVehicle(ExitTaxiVehicle);
        ExitAIDriveToJob.driveToTask:setTargetDirection(0, 0);
        ExitAIDriveToJob.driveToTask:setTargetPosition(-100,-100);

        g_currentMission.aiSystem:startJob(ExitAIDriveToJob, -1);
    end

    if TaxiState == 2 then

        g_currentMission:addExtraPrintText(g_i18n:getText("text_TAXISOON"));

        local Vx, Vy, Vz = getTranslation(TaxiVehicle.rootNode)
        local Distance = TaxiService:DistanceFrom(Vx,Vz,DestinationX,DestinationZ);

        if Distance < 25 or Input.isKeyPressed(Input.KEY_g) then 
            if AIDriveToJob.isRunning then
                g_currentMission.aiSystem:stopJobInternal(AIDriveToJob, 0);
            end

            g_currentMission:removeVehicle(TaxiVehicle);
            TaxiVehicle = nil;
            AIDriveToJob = nil;

            TaxiState = 3;

            local StartPosition = {
                x = Vx,
                y = Vy,
                z = Vz
            }

            TaxiService:SpawnTaxiVehicle(StartPosition,true);
            
            return;
        end


    end


    if TaxiState == 1 and AIDriveToJob ~= nil and g_currentMission.player.isControlled == false and g_currentMission.controlledVehicle == nil then 
        TaxiService:RequestLocation();
	end;

    if TaxiState == 0 and ExitTaxiVehicle ~= nil then

        local x, y, z = getTranslation(ExitTaxiVehicle.rootNode)
        local Px, Py, Pz = getTranslation(g_currentMission.player.rootNode);
        local Distance = TaxiService:DistanceFrom(x,z,Px,Pz);

        if Distance > 100 then 
            if ExitAIDriveToJob.isRunning then
                g_currentMission.aiSystem:stopJobInternal(ExitAIDriveToJob, 0);
            end
            g_currentMission:removeVehicle(ExitTaxiVehicle);
            ExitTaxiVehicle = nil;

            ExitAIDriveToJob = nil;
        end
	end;

    if TaxiState == 1 then

        if TaxiVehicle ~= nil then

            local x, y, z = getTranslation(TaxiVehicle.rootNode)
            local Distance = TaxiService:DistanceFrom(x,z,CoursePickupX,CoursePickupZ);

             if Distance > 1 then 
                 g_currentMission:addExtraPrintText(g_i18n:getText("text_TAXIARIVE").." ("..tostring(math.floor(Distance)).."m)");
            end

            if Distance < 1 then 
                g_currentMission:addExtraPrintText(g_i18n:getText("text_TAXIARIVECOMP"));
            end

        end

        if TaxiVehicle == nil then
            g_currentMission:addExtraPrintText(g_i18n:getText("text_TAXIARIVE"));
        end

	end;

    if TaxiState == 1 and TaxiVehicle ~= nil and AIDriveToJob ~= nil and not AIDriveToJob.isRunning then

        if StuckTrys > 50 then
            --print("UNREACHABLE");

            TaxiState = 0;

            if AIDriveToJob.isRunning then
                g_currentMission.aiSystem:stopJobInternal(AIDriveToJob, 0);
            end

            g_currentMission:removeVehicle(TaxiVehicle);

            CoursePickupX = nil;
            CoursePickupZ = nil;
            AIDriveToJob = nil;
            TaxiVehicle = nil;

            g_gui:showInfoDialog({text=g_i18n:getText("text_TAXIUNREACHABLE")})
            return;
        end



        local x, y, z = getTranslation(TaxiVehicle.rootNode)  
        local Distance = TaxiService:DistanceFrom(x,z,CoursePickupX,CoursePickupZ);

        LastDistanceToDestination = Distance;

        if Distance > 2 then 

        local NewStartX = x + math.random(-15, 15);
        local NewStartZ = z + math.random(-15, 15);
        local NewStartY = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, NewStartX, y, NewStartZ) + 1;

        local NewStartPosition = {
            x = NewStartX,
            y = NewStartY,
            z = NewStartZ
        }

        g_currentMission:removeVehicle(TaxiVehicle);
        TaxiService:SpawnTaxiVehicle(NewStartPosition);
        AIDriveToJob = nil;
        TaxiVehicle = nil;
        StuckTrys = StuckTrys + 1;
        --print("STUCK");

        end
    end;

	if TaxiState == 1 and TaxiVehicle ~= nil and AIDriveToJob == nil then
		AIDriveToJob = g_currentMission.aiJobTypeManager:createJob(g_currentMission.aiJobTypeManager:getJobTypeIndexByName("GOTO"))
		
        AIDriveToJob.vehicleParameter:setVehicle(TaxiVehicle);
		AIDriveToJob.positionAngleParameter:setAngle(0);
		AIDriveToJob.positionAngleParameter:setPosition(CoursePickupX,CoursePickupZ);

        AIDriveToJob.driveToTask:setVehicle(TaxiVehicle);
        AIDriveToJob.driveToTask:setTargetDirection(0, 0);
        AIDriveToJob.driveToTask:setTargetPosition(CoursePickupX,CoursePickupZ);


        g_currentMission.aiSystem:startJob(AIDriveToJob, -1);
	end;
end;

function TaxiService:VehicleCameraPostLoad()
	AllVehicleCameras[table.getn(AllVehicleCameras) + 1] = self;
end;

function TaxiService:GetSpawnPosition()

    local Px, Py, Pz = getTranslation(g_currentMission.player.rootNode);

    local StartPosition = {
        x = Px+50,
        y = Py,
        z = Pz+50
    }

    return StartPosition;
end;

function TaxiService:loadMap()
    Player.registerActionEvents = Utils.appendedFunction(Player.registerActionEvents, TaxiService.RegisterActionEvents)
    Player.updateActionEvents = Utils.appendedFunction(Player.updateActionEvents, TaxiService.UpdateActionEvents)
end;

local RequestTaxiEventId = nil

function TaxiService:RegisterActionEvents()
    TaxiService.events = {}
	local valid1, eventId1 = g_inputBinding:registerActionEvent(InputAction.REQUESTTAXI,"REQUESTTAXI",TaxiService.TaxiRequestEvent,false,true,false,true)
    if valid1 then
        table.insert(TaxiService.events, eventId1);
        g_inputBinding:setActionEventTextVisibility(eventId1, true);
        RequestTaxiEventId = eventId1;
    end
end

function TaxiService:UpdateActionEvents()
    
    g_inputBinding:setActionEventActive(RequestTaxiEventId, not g_currentMission.controlledVehicle);
    
    if g_currentMission.player.isControlled then

        if TaxiState == 0 then
            g_inputBinding:setActionEventText(RequestTaxiEventId, g_i18n:getText("input_REQUESTTAXI"))
        end

        if TaxiState ~= 0 then
            g_inputBinding:setActionEventText(RequestTaxiEventId, g_i18n:getText("input_CANCELTAXI"))
        end
    end
end

function TaxiService:TaxiRequestEvent(actionName, keyStatus, arg3, arg4, arg5)

    if TaxiState == 0 then

    TaxiState = 1;
    StuckTrys = 0;

    local Px, Py, Pz = getTranslation(g_currentMission.player.rootNode);
    CoursePickupX = Px+2;
    CoursePickupZ = Pz+2;

    if ExitTaxiVehicle ~= nil then 
        if ExitAIDriveToJob.isRunning then
            g_currentMission.aiSystem:stopJobInternal(ExitAIDriveToJob, 0);
        end

        TaxiVehicle = ExitTaxiVehicle;

        ExitTaxiVehicle = nil;
        ExitAIDriveToJob = nil;
        return;
    end;

    local StartPosition = TaxiService:GetSpawnPosition();
    TaxiService:SpawnTaxiVehicle(StartPosition);
    return;
    end

    if TaxiState ~= 0 and TaxiState ~= 2 and TaxiState ~= 3 then
        
        TaxiState = 0;

        if AIDriveToJob.isRunning then
            g_currentMission.aiSystem:stopJobInternal(AIDriveToJob, 0);
        end

        ExitTaxiVehicle = TaxiVehicle;


        CoursePickupX = nil;
        CoursePickupZ = nil;
        AIDriveToJob = nil;
        TaxiVehicle = nil;

        ExitAIDriveToJob = g_currentMission.aiJobTypeManager:createJob(g_currentMission.aiJobTypeManager:getJobTypeIndexByName("GOTO"))
		
        ExitAIDriveToJob.vehicleParameter:setVehicle(ExitTaxiVehicle);
		ExitAIDriveToJob.positionAngleParameter:setAngle(0);
		ExitAIDriveToJob.positionAngleParameter:setPosition(-100,-100);

        ExitAIDriveToJob.driveToTask:setVehicle(ExitTaxiVehicle);
        ExitAIDriveToJob.driveToTask:setTargetDirection(0, 0);
        ExitAIDriveToJob.driveToTask:setTargetPosition(-100,-100);

        g_currentMission.aiSystem:startJob(ExitAIDriveToJob, -1);
        return;
    end
end

VehicleCamera.onPostLoad = Utils.appendedFunction(VehicleCamera.onPostLoad,TaxiService.VehicleCameraPostLoad);
addModEventListener(TaxiService);